/***************************************************************************
****************************************************************************

This is a part of the xModule Source code.

Copyright (C) 2003-2009 The LS-Universe Team

  Homepage: http://www.ls-universe.info

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************
****************************************************************************/

/***************************************************************************
****************************************************************************

  Original source -> BadJerry <-

  Link: http://www.codeproject.com/gdi/htmldrawer.asp

  Completely Redone by Andymon

****************************************************************************
****************************************************************************/

#include "stdafx.h"
#include "HTMLDrawer.h"

#include "HTMLAtom.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CHTMLDrawer::CHTMLDrawer()
{
}

CHTMLDrawer::~CHTMLDrawer()
{
}

/////////////////////////////////////////////////////////////
// DrawText : static implementation
/////////////////////////////////////////////////////////////
// Parameters:
// pDC : Device context
// strText: Text to be drawn
// fntDefault : Default font to be used
// rctPosition : Rectangle in which the text is to be drawn
// nAlignment: 0 (left) or 1 (centered) or 2 (right)
/////////////////////////////////////////////////////////////

void CHTMLDrawer::DrawText(CDC * pDC, const CString & strText,const CHTMLFont & fntDefault,CRect & rctPosition,short nAlignment,short nVAlignment, bool lineBreak, int newlineSpace, bool scrolling, StringList &listHREFClickRegions, bool noFontColorChange)
{
    CObArray arrDisplayAtoms;
    CObArray arrAnchors;
        
    // Find where each item is to be drawn
    FindDisplayAtoms(pDC,strText,fntDefault,rctPosition,nAlignment,nVAlignment,lineBreak,newlineSpace,arrDisplayAtoms,arrAnchors,noFontColorChange);

    // Find href tags for label export if not
    if ( arrAnchors.GetSize() != 0 )
        ConvertAnchorInStringList(arrAnchors, listHREFClickRegions);

    // Draw them
    DrawAtoms(pDC,arrDisplayAtoms,rctPosition,scrolling);

    // Free memory
    CHTMLAtom::DeleteArray(arrDisplayAtoms);
}

POINT CHTMLDrawer::MeasureText(CDC * pDC, const CString & strText,const CHTMLFont & fntDefault,CRect & rctPosition,short nAlignment,short nVAlignment, bool lineBreak, int newlineSpace)
{
    POINT returnValues;
    int maxWidth = 0;
    int maxHeight = 0;

    CObArray arrDisplayAtoms;
    CObArray arrAnchors;

    FindDisplayAtoms(pDC, strText,fntDefault,rctPosition,nAlignment,nVAlignment,lineBreak,newlineSpace,arrDisplayAtoms,arrAnchors,false);

    for ( int nAtom = 0; nAtom < arrDisplayAtoms.GetSize(); nAtom++ )
    {
        CHTMLAtom * pAtom = (CHTMLAtom *) arrDisplayAtoms.GetAt(nAtom);
    
        if (nAlignment == 0 || nAlignment == 1)
        {
            if (maxWidth < pAtom->GetPosition().right)
                maxWidth = pAtom->GetPosition().right;
        }
        else
        {
            if (maxWidth > pAtom->GetPosition().left)
                maxWidth = pAtom->GetPosition().left;
        }

        if (nVAlignment == 0 || nVAlignment == 1)
        {
            if (maxHeight < pAtom->GetPosition().bottom)
                maxHeight = pAtom->GetPosition().bottom;
        }
        else
        {
            if (maxHeight > pAtom->GetPosition().top)
                maxHeight = pAtom->GetPosition().top;
        }
    }

    // Free memory
    CHTMLAtom::DeleteArray(arrDisplayAtoms);

    if (nAlignment == 1)
        maxWidth = maxWidth*2;
    else if (nAlignment == 2)
        maxWidth = -maxWidth;

    if (nVAlignment == 1)
        maxHeight = maxHeight*2;
    else if (nVAlignment == 2)
        maxHeight = -maxHeight;

    returnValues.x = maxWidth;
    returnValues.y = maxHeight;

    return returnValues;
}

int CHTMLDrawer::FindDisplayAtoms(CDC * pDC, const CString & strText,const CHTMLFont & fntDefault,CRect & rctPosition,short nAlignment,short nVAlignment, bool lineBreak, int newlineSpace, CObArray & arrDisplayAtoms,CObArray & arrAnchors,bool noFontColorChange)
{
    // Parse the text to break into atoms having the same settings
    CObArray arrAtoms;

    ParseHTMLText(strText,fntDefault,arrAtoms,nAlignment,noFontColorChange);

    // If no text return
    if ( arrAtoms.GetSize() == 0 )
        return 0;

    // Break into smaller atoms according to end of line
    BreakIntoLines(pDC,arrAtoms,arrDisplayAtoms,rctPosition,arrAnchors,lineBreak,newlineSpace);

    CHTMLAtom::DeleteArray(arrAtoms);

    // Find the lowest position to center vertically
    int nLowest = 0;
    int nAscent = 0;
    int nHeight = 0;
    for ( int nAtom = 0; nAtom < arrDisplayAtoms.GetSize(); nAtom++ )
    {
        CHTMLAtom * pAtom = (CHTMLAtom *) arrDisplayAtoms.GetAt(nAtom);
        if ( pAtom->GetPosition().bottom > nLowest )
            nLowest = pAtom->GetPosition().bottom;
        if ( pAtom->GetAscent() > nAscent )
        {
            nAscent = pAtom->GetAscent();
            nHeight = pAtom->GetAdrPosition()->Height();
        }
    }

    int nShiftVertically = 0;

    //Vertical Alignment
    //Top
    if (nVAlignment == 0)
        nShiftVertically = -(nHeight-nAscent);
    //Center
    else if (nVAlignment == 1)
        nShiftVertically = ((rctPosition.Height()-nLowest)-(nHeight-nAscent)) / 2;
    //Bottom
    else if (nVAlignment == 2)
        nShiftVertically = rctPosition.Height()-nLowest;

    ApplyAlignments(arrDisplayAtoms,rctPosition.top+nShiftVertically,rctPosition.right);

    return nLowest+nShiftVertically;
}


void CHTMLDrawer::ParseHTMLText(const CString & strSource,const CHTMLFont & fntDefault,CObArray & arrAtoms,short nDefaultAlignment,bool noFontColorChange)
{
    CString strText(strSource);
    CHTMLFont fntCurrent(fntDefault);

    CObArray arrFontStack;
    int nTagStart;
    short nAlignment = nDefaultAlignment;
    bool isLastAtomNewLine = true;
    CString strURL = "";
    CString strTooltip = "";
    while ( (nTagStart = strText.Find('<')) != -1 )
    {
        int nTagEnd = strText.Find('>',nTagStart);
        if ( nTagEnd == -1 )
            break;
        
        if ( AddText(arrAtoms,strText.Left(nTagStart),fntCurrent,nAlignment,strURL,strTooltip) )
            isLastAtomNewLine = false;

        CString strTag = strText.Mid(nTagStart+1,nTagEnd-nTagStart-1);
        bool isNewLine = false;
        UpdatePoliceAccordingToTag(strTag,fntCurrent,arrFontStack,isNewLine,nAlignment,nDefaultAlignment,isLastAtomNewLine,strURL,strTooltip,noFontColorChange);
        strText = strText.Mid(nTagEnd+1);
        if ( isNewLine )
        {
            AddText(arrAtoms,_T("\r\n"),fntCurrent,nAlignment,strURL,strTooltip);
            isLastAtomNewLine = true;
        }
    }
    if ( !strText.IsEmpty() )
        AddText(arrAtoms,strText,fntCurrent,nAlignment,strURL,strTooltip);

    CHTMLFont::DeleteArray(arrFontStack);
}

bool CHTMLDrawer::AddText(CObArray & arrAtoms,const CString & strCodedText,CHTMLFont & polCurrent,short nAlignment,const CString & strURL,const CString & strTooltip)
{
    if ( strCodedText.IsEmpty() )
        return false;

    // Translate &quot; &lt; &gt; and &amp;
    CString strText(strCodedText);

    int nTagStart = 0;
    while ( (nTagStart = strText.Find('&', nTagStart)) != -1 )
    {
        int nTagEnd = strText.Find(';',nTagStart);
    
        if (nTagEnd == -1)
            break;

        CString strEnd=strText.Mid(nTagEnd+1);
        CString strTag=strText.Mid(nTagStart+1,nTagEnd-nTagStart-1);
        strText=strText.Left(nTagStart);
        if ( strTag.CompareNoCase(_T("quot")) == 0 )
            strText+='\"';
        else if ( strTag.CompareNoCase(_T("lt")) == 0 )
            strText+='<';
        else if ( strTag.CompareNoCase(_T("gt")) == 0 )
            strText+='>';
        else if ( strTag.CompareNoCase(_T("amp")) == 0 )
            strText+='&';
        else
        {
            strText+="&";
            strText+=strTag;
            if ( nTagEnd != -1)
                strText+=";";
        }
        strText+=strEnd;
        nTagStart++;
    }

    CHTMLAtom * pAtom=new CHTMLAtom;
    pAtom->SetText(strText);
    pAtom->SetHTMLFont(polCurrent);
    pAtom->SetAlignment(nAlignment);
    pAtom->SetURL(strURL);
    pAtom->SetTooltip(strTooltip);
    arrAtoms.Add(pAtom);

    return true;
}

void CHTMLDrawer::UpdatePoliceAccordingToTag(CString & strTag,CHTMLFont & polCurrent,CObArray & arrFontStack,bool & isNewLine,short & nAlignment,short nDefaultAlignment,bool isLastAtomNewLine,CString & strURL,CString & strTooltip, bool noFontColorChange)
{
    strTag.TrimLeft();
    strTag.TrimRight();
    if ( strTag.CompareNoCase(_T("b")) == 0 )
        polCurrent.SetBold(true);
    else if ( strTag.CompareNoCase(_T("strong")) == 0 )
        polCurrent.SetBold(true);
    else if ( strTag.CompareNoCase(_T("/b")) == 0 )
        polCurrent.SetBold(false);
    else if ( strTag.CompareNoCase(_T("/strong")) == 0 )
        polCurrent.SetBold(false);
    else if ( strTag.CompareNoCase(_T("i")) == 0 )
        polCurrent.SetItalic(true);
    else if ( strTag.CompareNoCase(_T("/i")) == 0 )
        polCurrent.SetItalic(false);
    else if ( strTag.CompareNoCase(_T("em")) == 0 )
        polCurrent.SetItalic(true);
    else if ( strTag.CompareNoCase(_T("/em")) == 0 )
        polCurrent.SetItalic(false);
    else if ( strTag.CompareNoCase(_T("u")) == 0 )
        polCurrent.SetUnderline(true);
    else if ( strTag.CompareNoCase(_T("/u")) == 0 )
        polCurrent.SetUnderline(false);
    else if ( strTag.CompareNoCase(_T("br")) == 0 )
        isNewLine = true;
    else if ( strTag.Left(1).CompareNoCase(_T("p")) == 0 )
    {
        if ( !isLastAtomNewLine)
            isNewLine = true;

        if ( strTag.GetLength() == 1 )
            return;
        CStringArray arrNames,arrValues;
        FindTagAttributes(strTag.Mid(2),arrNames,arrValues);

        for ( int nAttribute = 0; nAttribute < arrNames.GetSize(); nAttribute++)
        {
            if ( arrNames.GetAt(nAttribute).CompareNoCase(_T("align")) == 0 )
            {
                CString strValue(arrValues.GetAt(nAttribute));
                if ( strValue.CompareNoCase(_T("left")) == 0 )
                    nAlignment=0;
                else if ( strValue.CompareNoCase(_T("center")) == 0 )
                    nAlignment=1;
                else if ( strValue.CompareNoCase(_T("right")) == 0 )
                    nAlignment=2;
            }
        }
    }
    else if ( strTag.CompareNoCase(_T("/p")) == 0 )
    {
        if ( !isLastAtomNewLine)
            isNewLine = true;
        nAlignment=nDefaultAlignment;
    }
    else if ( strTag.CompareNoCase(_T("center")) == 0 )
    {
        if ( !isLastAtomNewLine)
            isNewLine = true;
        nAlignment=1;
    }
    else if ( strTag.CompareNoCase(_T("/center")) == 0 )
    {
        if ( !isLastAtomNewLine)
            isNewLine = true;
        nAlignment=nDefaultAlignment;
    }
    else if ( strTag.Left(1).CompareNoCase(_T("a")) == 0 )
    {
        if ( strTag.GetLength() == 1 )
            return;
        CStringArray arrNames,arrValues;
        FindTagAttributes(strTag.Mid(2),arrNames,arrValues);
        
        for ( int nAttribute = 0; nAttribute < arrNames.GetSize(); nAttribute++)
        {
            if ( arrNames.GetAt(nAttribute).CompareNoCase(_T("href")) == 0 )
                strURL = arrValues.GetAt(nAttribute);
            
            if ( arrNames.GetAt(nAttribute).CompareNoCase(_T("title")) == 0 )
                strTooltip = arrValues.GetAt(nAttribute);
        }
    }
    else if ( strTag.Left(2).CompareNoCase(_T("/a")) == 0 )
    {
        strURL.Empty();
        strTooltip.Empty();
    }
    else if ( strTag.Left(4).CompareNoCase(_T("font")) == 0 )
    {
        // Store previous police and color in stack
        arrFontStack.Add((CObject *) new CHTMLFont(polCurrent));

        CStringArray arrNames,arrValues;
        FindTagAttributes(strTag.Mid(4),arrNames,arrValues);

        for ( int nAttribute = 0; nAttribute < arrNames.GetSize(); nAttribute++)
        {
            CString strValue(arrValues.GetAt(nAttribute));
            if ( !strValue.IsEmpty() )
            {
                CString strName(arrNames.GetAt(nAttribute));

                // Update current police
                if ( strName.CompareNoCase(_T("face")) == 0 )
                {
                    polCurrent.SetName(strValue);
                }
                else if ( strName.CompareNoCase(_T("size")) == 0 )
                {
                    int nSize = ToLong(strValue);
                
                    polCurrent.SetSize(nSize);
                }
                else if ( !noFontColorChange && strName.CompareNoCase(_T("color")) == 0 )
                {
                    if ( strValue[0] == '#' )
                    {
                        strValue = strValue.Mid(1);
                        if ( strValue.GetLength() < 6 )
                        {
                            CString strTemp;
                            strTemp.Format(_T("%0*d%s"),6-strValue.GetLength(),0,strValue);
                            strValue = strTemp;
                        }
                        polCurrent.SetColor(RGB(FromHex(strValue.Left(2)),FromHex(strValue.Mid(2,2)),FromHex(strValue.Mid(4,2))));
                    }
                }
                else if ( noFontColorChange && (strName.CompareNoCase(_T("shadowcolor")) == 0 || strName.CompareNoCase(_T("outlinecolor")) == 0) )
                {
                    if ( strValue[0] == '#' )
                    {
                        strValue = strValue.Mid(1);
                        if ( strValue.GetLength() < 6 )
                        {
                            CString strTemp;
                            strTemp.Format(_T("%0*d%s"),6-strValue.GetLength(),0,strValue);
                            strValue = strTemp;
                        }
                        polCurrent.SetColor(RGB(FromHex(strValue.Left(2)),FromHex(strValue.Mid(2,2)),FromHex(strValue.Mid(4,2))));
                    }
                }
            }
        }
    }
    else if ( strTag.Left(5).CompareNoCase(_T("/font")) == 0 )
    {
        if ( arrFontStack.GetSize() >0 )
        {
            CHTMLFont * pFont = (CHTMLFont *) arrFontStack.GetAt(arrFontStack.GetSize()-1);
            polCurrent = *pFont;
            delete pFont;
            arrFontStack.SetSize(arrFontStack.GetSize()-1);
        }
    }
}

void CHTMLDrawer::FindTagAttributes(const CString & strAttributes,CStringArray & arrNames,CStringArray & arrValues)
{
    CString strInfo = strAttributes;
    while ( true )
    {
        int nPos = strInfo.Find(_T('='));
        if ( nPos == -1 )
            break;

        // Find the name of the property
        CString strName = strInfo.Left(nPos);
        strName.TrimLeft();
        strName.TrimRight();

        // Find the start of the value
        strInfo = strInfo.Mid(nPos+1);
        while ( strInfo[0] == ' ')
            strInfo = strInfo.Mid(1);

        // Find the value
        CString strValue;
        if ( strInfo[0] == '\'' )
        {
            nPos = strInfo.Find('\'',1);
            if ( nPos != -1 )
            {
                strValue=strInfo.Mid(1,nPos-1);
                strInfo = strInfo.Mid(nPos+1);
            }
        }
        else
        {
            nPos = strInfo.Find(' ');
            if ( nPos != -1 )
            {
                strValue=strInfo.Left(nPos);
                strInfo = strInfo.Mid(nPos+1);
            }
            else
            {
                strValue=strInfo;
                strInfo.Empty();
            }
        }

        arrNames.Add(strName);
        arrValues.Add(strValue);
    }
}


void CHTMLDrawer::BreakIntoLines(CDC * pDC,CObArray & arrAtoms,CObArray & arrDisplayAtoms,CRect & rctPosition,CObArray & arrAnchors, bool lineBreak, int newlineSpace)
{
    // Select the font
    bool useAscentAlignment = false;
    CFont* pCurrentFont = NULL;
    CFont* pOldFont=NULL;

    int nY = 0;
    int nX = rctPosition.left;
    int nXMax = rctPosition.right;
    int nStartLineAtom = 0;
    int nLowestAscent = 0;
    int nLine = 0;

    for ( int nAtom = 0; nAtom < arrAtoms.GetSize(); nAtom++)
    {
        CHTMLAtom * pAtom = (CHTMLAtom *) arrAtoms.GetAt(nAtom);

        // Find if there is a line break in the text
        bool isLineBreak = false;
        int nPos = pAtom->GetText().Find(_T("\r\n"));
        if ( nPos != -1 )
        {
            // Split this atom into 2
            CString strTextBefore = pAtom->GetText().Left(nPos);
            CString strTextAfter = pAtom->GetText().Mid(nPos+2);
            
            if ( !strTextAfter.IsEmpty() )
            {
                CHTMLAtom * pNewAtom = new CHTMLAtom(*pAtom);
                pNewAtom->SetText(strTextAfter);
                arrAtoms.InsertAt(nAtom+1,pNewAtom);
            }
            pAtom->SetText(strTextBefore);

            isLineBreak = true;
        }

        // Select the current font
        if ( pCurrentFont == NULL )
        {
            pCurrentFont = pAtom->GetHTMLFont().GetFont(pDC);
            pOldFont = pDC->SelectObject(pCurrentFont);
        }
        else
        {
            CFont * pNewFont = pAtom->GetHTMLFont().GetFont(pDC);
            pDC->SelectObject(pNewFont);
            delete pCurrentFont;
            pCurrentFont = pNewFont;
        }

        // Find the size of the text
        CString strTextToWrite = pAtom->GetText();
        while ( !strTextToWrite.IsEmpty() )
        {
            bool isCreateNewLine = false;
            CSize sizText = pDC->GetTextExtent(strTextToWrite);

            // If it does not fit in the current line
            CString strPart = strTextToWrite;

            if (lineBreak)
            {
                while ( nX + sizText.cx >= nXMax)
                {
                    RemoveLastWord(strPart);
                    sizText = pDC->GetTextExtent(strPart);
                    isCreateNewLine = true;
                    if ( strPart.IsEmpty() )
                        break;
                }

                // If nothing fits, 
                if ( strPart.IsEmpty() )
                {
                    // If the current line is empty
                    if ( nStartLineAtom == arrDisplayAtoms.GetSize() )
                    {
                        // We have to break up anywhere
                        strPart=strTextToWrite[0];
                        for ( int nChar = 1; nChar < strTextToWrite.GetLength(); nChar++)
                        {
                            strPart+=strTextToWrite[nChar];
                            sizText = pDC->GetTextExtent(strPart);
                            if ( nX + sizText.cx > nXMax)
                            {
                                strPart = strPart.Left(strPart.GetLength()-1);
                                break;
                            }
                        }
                    }
                }
            }

            if ( !strPart.IsEmpty() )
            {
                // We can draw strPart
                sizText = pDC->GetTextExtent(strPart);
                CRect rctTextPosition(nX,nY,nX+sizText.cx,nY+sizText.cy);

                // Get the text metrics so we can align text of different sizes
                TEXTMETRIC tmMetrics;
                pDC->GetTextMetrics(&tmMetrics);

                CHTMLAtom * pDisplayAtom = new CHTMLAtom(*pAtom);
                pDisplayAtom->SetText(strPart);
                pDisplayAtom->SetPosition(rctTextPosition);
                pDisplayAtom->SetAscent(tmMetrics.tmAscent);
                pDisplayAtom->SetLine(nLine);
                arrDisplayAtoms.Add(pDisplayAtom);

                if ( !pDisplayAtom->GetURL().IsEmpty() )
                    arrAnchors.Add(pDisplayAtom);

                nX = rctTextPosition.right;
                strTextToWrite = strTextToWrite.Mid(strPart.GetLength());

                // Find the lowest position on the line to align properly
                if ( tmMetrics.tmAscent > nLowestAscent )
                    nLowestAscent = sizText.cy;
            }

            if (lineBreak)
            {
                if ( isCreateNewLine )
                {
                    // Align to lowest position
                    AlignLineAccordingToAscent(arrDisplayAtoms,nLowestAscent,nStartLineAtom);

                    CreateNewLine(pDC,arrDisplayAtoms,nStartLineAtom,nY,newlineSpace);
                    nX=rctPosition.left;
                    strTextToWrite.TrimLeft();
                    nLine++;
                }
            }
        }
        
        if ( isLineBreak )
        {
            // Align to lowest position
            AlignLineAccordingToAscent(arrDisplayAtoms,nLowestAscent,nStartLineAtom);
            nLowestAscent=0;

            CreateNewLine(pDC,arrDisplayAtoms,nStartLineAtom,nY,newlineSpace);
            nX=rctPosition.left;
            nLine++;
        }
    }

    // Align to lowest position
    AlignLineAccordingToAscent(arrDisplayAtoms,nLowestAscent,nStartLineAtom);
    nLowestAscent=0;

    if ( pOldFont )
        pDC->SelectObject(pOldFont);
    delete pCurrentFont;
}

void CHTMLDrawer::AlignLineAccordingToAscent(CObArray & arrDisplayAtoms,int nLowestAscent,int nStartLineAtom)
{
    for ( int nAtom = nStartLineAtom; nAtom < arrDisplayAtoms.GetSize(); nAtom++ )
    {
        CHTMLAtom * pAtom = (CHTMLAtom *) arrDisplayAtoms.GetAt(nAtom);
        if ( pAtom->GetAscent() != nLowestAscent )
        {
            CRect * pRect = pAtom->GetAdrPosition();
            *pRect+= CPoint(0,nLowestAscent-pAtom->GetAscent());
        }
    }
}


void CHTMLDrawer::CreateNewLine(CDC * pDC,CObArray & arrDisplayAtoms,int & nStartLineAtom,int & nY, int newlineSpace)
{
    // If no atom has been created for this line, create a dummy empty atom 
    if ( nStartLineAtom == arrDisplayAtoms.GetSize() )
    {
        CString strText=_T(" ");
        CSize sizText = pDC->GetTextExtent(strText);
        CRect rctTextPosition(0,nY,0+sizText.cx,nY+sizText.cy);

        CHTMLAtom * pDisplayAtom = new CHTMLAtom;
        pDisplayAtom->SetText(strText);
        pDisplayAtom->SetPosition(rctTextPosition);
        arrDisplayAtoms.Add(pDisplayAtom);

        nY=rctTextPosition.bottom + newlineSpace;
        nStartLineAtom = arrDisplayAtoms.GetSize();

        return;
    }

    // Find the lowest atom on this line
    int nMaxBottom = nY;
    for ( int nAtom = nStartLineAtom; nAtom < arrDisplayAtoms.GetSize(); nAtom++)
    {
        CHTMLAtom * pLineAtom=(CHTMLAtom *) arrDisplayAtoms.GetAt(nAtom);
        if ( pLineAtom->GetPosition().bottom > nMaxBottom )
        {
            nMaxBottom = pLineAtom->GetPosition().bottom;
        }
    }
    /*
    // Align them to the lowest
    for ( nAtom = nStartLineAtom; nAtom < arrDisplayAtoms.GetSize(); nAtom++)
    {
        CHTMLAtom * pLineAtom=(CHTMLAtom *) arrDisplayAtoms.GetAt(nAtom);
        CRect rctPosition = pLineAtom->GetPosition();
        int nShift = nMaxBottom - rctPosition.bottom;
        rctPosition.top+=nShift;
        rctPosition.bottom+=nShift;
        pLineAtom->SetPosition(rctPosition);
    }
    */
    nY=nMaxBottom + newlineSpace;
    nStartLineAtom = arrDisplayAtoms.GetSize();
}

void CHTMLDrawer::RemoveLastWord(CString & strText)
{
    bool isOneCharRemoved = false;
    while ( strText.GetLength() != 0 )
    {
        TCHAR car = strText[strText.GetLength()-1];
        //if ( ( ( car < 'a' ) || ( car > 'z' ) ) && ( ( car < 'A' ) || ( car > 'Z' ) ) )
        if ( car == ' ' || car == '-' )
        {
            if (!isOneCharRemoved )
                strText = strText.Left(strText.GetLength()-1);
            break;
        }
        strText = strText.Left(strText.GetLength()-1);
        isOneCharRemoved = true;
    }
}

void CHTMLDrawer::ApplyAlignments(CObArray & arrDisplayAtoms,int nShiftVertically,int nRight)
{
    int nLine=-1;
    int nShiftX = 0;
    for ( int nAtom = 0; nAtom < arrDisplayAtoms.GetSize(); nAtom++)
    {
        CHTMLAtom * pAtom = (CHTMLAtom *) arrDisplayAtoms.GetAt(nAtom);

        // If we have changed line
        if ( nLine != pAtom->GetLine() )
        {
            nLine = pAtom->GetLine();
            // If there is an alignent for this line
            if ( pAtom->GetAlignment() == 0 )
            {
                nShiftX = 0;
            }
            else
            {
                // Find the extent of the line
                int nMaxRight = pAtom->GetPosition().right;
                for ( int nFollowingAtom = nAtom+1; nFollowingAtom < arrDisplayAtoms.GetSize(); nFollowingAtom++)
                {
                    CHTMLAtom * pFollowingAtom = (CHTMLAtom *) arrDisplayAtoms.GetAt(nFollowingAtom);
                    if ( nLine != pFollowingAtom->GetLine() )
                        break;
                    nMaxRight = pFollowingAtom->GetPosition().right;
                }
                if (pAtom->GetAlignment() == 1) 
                    nShiftX = (nRight - nMaxRight)/2;
                else
                    nShiftX = nRight - nMaxRight-1;
            }
        }

        // Change the position of the atom
        *pAtom->GetAdrPosition() += CPoint(nShiftX,nShiftVertically);
    }
}

void CHTMLDrawer::DrawAtoms(CDC * pDC,CObArray & arrDisplayAtoms, CRect & rctPosition, bool scrolling)
{
    // Prepare the settings to be restored
    CFont* pCurrentFont = NULL;
    CFont* pOldFont;
    COLORREF rgbOldColor;
    CRect actualDrawRect;
    CString actualText;
    pDC->SetBkMode(TRANSPARENT);

    for ( int nAtom = 0; nAtom < arrDisplayAtoms.GetSize(); nAtom++)
    {
        CHTMLAtom * pAtom = (CHTMLAtom *) arrDisplayAtoms.GetAt(nAtom);

        CHTMLFont htmlFont(pAtom->GetHTMLFont());

        // Select the font
        if ( pCurrentFont == NULL )
        {
            pCurrentFont = htmlFont.GetFont(pDC);
            pOldFont = pDC->SelectObject(pCurrentFont);
            rgbOldColor  = pDC->SetTextColor(htmlFont.GetColor());
        }
        else
        {
            CFont * pNewFont = htmlFont.GetFont(pDC);
            pDC->SelectObject(pNewFont);
            delete pCurrentFont;
            pCurrentFont = pNewFont;
            pDC->SetTextColor(htmlFont.GetColor());
        }

        actualDrawRect.CopyRect(pAtom->GetAdrPosition());
        actualText = pAtom->GetText();

        if ( !scrolling )
        {
            int nDifference = pAtom->GetAdrPosition()->right - pAtom->GetAdrPosition()->Width();
            SIZE szStr;
            CString strTemp;
            int nWidth;

            if ( nDifference < rctPosition.left )
            {   
                strTemp = pAtom->GetText();
                nWidth = pAtom->GetAdrPosition()->Width() - (rctPosition.left-nDifference);
                
                //
                // Check if the text extent is larger than the rect's width
                //
                //
                // Insert the initial ellipsis, by replacing the first char
                //
                strTemp.Insert(0, "...");

                GetTextExtentPoint32(pDC->m_hDC, strTemp, strTemp.GetLength(), &szStr);
                if(szStr.cx > nWidth)
                {
                    //
                    // While the extents are larger than the width, remove one
                    // character at the time
                    //
                    while(szStr.cx > nWidth)
                    {
                        strTemp.Delete(3,1);
    
                        GetTextExtentPoint32(pDC->m_hDC, strTemp, strTemp.GetLength(), &szStr);
                    }
                }

                actualText = strTemp;
                actualDrawRect.left = rctPosition.left;
            }

            nDifference = actualDrawRect.left + actualDrawRect.Width();

            if ( nDifference > rctPosition.right )
            {   
                strTemp = actualText;
                nWidth = rctPosition.right - actualDrawRect.left;
                
                //
                // Check if the text extent is larger than the rect's width
                //
                //
                // Insert the initial ellipsis, by adding
                //
                strTemp += "...";

                GetTextExtentPoint32(pDC->m_hDC, strTemp, strTemp.GetLength(), &szStr);
                if(szStr.cx > nWidth)
                {
                    //
                    // While the extents are larger than the width, remove one
                    // character at the time
                    //
                    while(szStr.cx > nWidth)
                    {
                        strTemp.Delete(strTemp.GetLength()-4,1);
    
                        GetTextExtentPoint32(pDC->m_hDC, strTemp, strTemp.GetLength(), &szStr);
                    }
                }

                actualText = strTemp;
                actualDrawRect.right = rctPosition.right;
            }
        }

        // Draw the text
        pDC->DrawText(actualText/*pAtom->GetText()*/,&actualDrawRect/*pAtom->GetAdrPosition()*/, DT_LEFT | DT_NOPREFIX | DT_SINGLELINE | DT_NOCLIP);
    }

    // Restore the settings
    pDC->SelectObject(pOldFont);
    pDC->SetTextColor(rgbOldColor);
    delete pCurrentFont;
}


long CHTMLDrawer::FromHex(const CString & strHex)
{
    CString strHexa(strHex);
    strHexa.TrimRight();
    strHexa.TrimLeft();
    long nRes=0;
    long nBase=1;
    for ( int nChar = strHexa.GetLength()-1; nChar >= 0; nChar--)
    {
        switch (strHexa[nChar])
        {
        case '0':
        case '1':
        case '2':
        case '3':
        case '4':
        case '5':
        case '6':
        case '7':
        case '8':
        case '9':
            nRes+=(strHexa[nChar]-'0')*nBase;
            break;
        case 'a':
        case 'b':
        case 'c':
        case 'd':
        case 'e':
        case 'f':
            nRes+=(10+strHexa[nChar]-'a')*nBase;
            break;
        case 'A':
        case 'B':
        case 'C':
        case 'D':
        case 'E':
        case 'F':
            nRes+=(10+strHexa[nChar]-'A')*nBase;
            break;
        }
        nBase*=16;
    }
    return nRes;
}


long CHTMLDrawer::ToLong(LPCTSTR lpszText)
{
    ASSERT(lpszText != NULL);

    while (*lpszText == ' ' || *lpszText == '\t')
        lpszText++;
    TCHAR chFirst = lpszText[0];
    long l;
    l = _tcstol(lpszText, (LPTSTR*)&lpszText, 10);

    if (l == 0 && chFirst != '0')
        return 0;   // could not convert

    while (*lpszText == ' ' || *lpszText == '\t')
        lpszText++;
    if (*lpszText != '\0')
        return 0;   // not terminated properly

    // all ok
    return l;
}

void CHTMLDrawer::ConvertAnchorInStringList(CObArray & m_arrAnchors, StringList &listHREFClickRegions)
{
    listHREFClickRegions.clear();
    for ( int nAnchor = 0; nAnchor < m_arrAnchors.GetSize(); nAnchor++)
    {
        CHTMLAtom * pAnchor = (CHTMLAtom *) m_arrAnchors.GetAt(nAnchor);

        CString region;
        region.Format("%i %i %i %i \"%s\"<'%s'>%s", pAnchor->GetPosition().left, pAnchor->GetPosition().top, pAnchor->GetPosition().right, pAnchor->GetPosition().bottom, pAnchor->GetTooltip(), pAnchor->GetURL(), pAnchor->GetText());

        listHREFClickRegions.insert(listHREFClickRegions.end(), (string)region);
    }
}